# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['datar',
 'datar.base',
 'datar.core',
 'datar.datar',
 'datar.datasets',
 'datar.dplyr',
 'datar.forcats',
 'datar.stats',
 'datar.tibble',
 'datar.tidyr',
 'datar.utils']

package_data = \
{'': ['*']}

install_requires = \
['diot>=0.1.1,<0.2.0', 'pandas>=1.2,<2.0', 'pipda>=0.4.5,<0.5.0']

setup_kwargs = {
    'name': 'datar',
    'version': '0.5.4',
    'description': 'Port of dplyr and other related R packages in python, using pipda.',
    'long_description': '# datar\n\nPort of [dplyr][2] and other related R packages in python, using [pipda][3].\n\n<!-- badges -->\n[![Pypi][6]][7] [![Github][8]][9] ![Building][10] [![Docs and API][11]][5] [![Codacy][12]][13] [![Codacy coverage][14]][13]\n\n[Documentation][5] | [Reference Maps][15] | [Notebook Examples][16] | [API][17] | [Blog][18]\n\n<img width="30%" style="margin: 10px 10px 10px 30px" align="right" src="logo.png">\n\nUnlike other similar packages in python that just mimic the piping syntax, `datar` follows the API designs from the original packages as much as possible, and is tested thoroughly with the cases from the original packages. So that minimal effort is needed for those who are familar with those R packages to transition to python.\n\n\n## Installtion\n\n```shell\npip install -U datar\n# to make sure dependencies to be up-to-date\n# pip install -U varname pipda datar\n```\n\n`datar` requires python 3.7.1+ and is backended by `pandas (1.2+)`.\n\n## Example usage\n\n```python\nfrom datar import f\nfrom datar.dplyr import mutate, filter, if_else\nfrom datar.tibble import tibble\n# or\n# from datar.all import f, mutate, filter, if_else, tibble\n\ndf = tibble(\n    x=range(4),\n    y=[\'zero\', \'one\', \'two\', \'three\']\n)\ndf >> mutate(z=f.x)\n"""# output\n        x        y       z\n  <int64> <object> <int64>\n0       0     zero       0\n1       1      one       1\n2       2      two       2\n3       3    three       3\n"""\n\ndf >> mutate(z=if_else(f.x>1, 1, 0))\n"""# output:\n        x        y       z\n  <int64> <object> <int64>\n0       0     zero       0\n1       1      one       0\n2       2      two       1\n3       3    three       1\n"""\n\ndf >> filter(f.x>1)\n"""# output:\n        x        y\n  <int64> <object>\n0       2      two\n1       3    three\n"""\n\ndf >> mutate(z=if_else(f.x>1, 1, 0)) >> filter(f.z==1)\n"""# output:\n        x        y       z\n  <int64> <object> <int64>\n0       2      two       1\n1       3    three       1\n"""\n```\n\n```python\n# works with plotnine\n# example grabbed from https://github.com/has2k1/plydata\nimport numpy\nfrom datar.base import sin, pi\nfrom plotnine import ggplot, aes, geom_line, theme_classic\n\ndf = tibble(x=numpy.linspace(0, 2*pi, 500))\n(df >>\n  mutate(y=sin(f.x), sign=if_else(f.y>=0, "positive", "negative")) >>\n  ggplot(aes(x=\'x\', y=\'y\')) +\n  theme_classic() +\n  geom_line(aes(color=\'sign\'), size=1.2))\n```\n\n![example](./example.png)\n\n```python\n# easy to integrate with other libraries\n# for example: klib\nimport klib\nfrom pipda import register_verb\nfrom datar.datasets import iris\nfrom datar.dplyr import pull\n\ndist_plot = register_verb(func=klib.dist_plot)\niris >> pull(f.Sepal_Length) >> dist_plot()\n```\n\n![example](./example2.png)\n\n## CLI interface\n\nSee [datar-cli][19]\n\nExample:\n```shell\n❯ datar import table2 | datar head\n       country    year        type      count\n      <object> <int64>    <object>    <int64>\n0  Afghanistan    1999       cases        745\n1  Afghanistan    1999  population   19987071\n2  Afghanistan    2000       cases       2666\n3  Afghanistan    2000  population   20595360\n4       Brazil    1999       cases      37737\n5       Brazil    1999  population  172006362\n```\n\n```shell\n❯ datar import table2 | \\\n    datar mutate --count "if_else(f.year==1999, f.count*2, f.count)"\n        country    year        type       count\n       <object> <int64>    <object>     <int64>\n0   Afghanistan    1999       cases        1490\n1   Afghanistan    1999  population    39974142\n2   Afghanistan    2000       cases        2666\n3   Afghanistan    2000  population    20595360\n4        Brazil    1999       cases       75474\n5        Brazil    1999  population   344012724\n6        Brazil    2000       cases       80488\n7        Brazil    2000  population   174504898\n8         China    1999       cases      424516\n9         China    1999  population  2545830544\n10        China    2000       cases      213766\n11        China    2000  population  1280428583\n```\n\n[1]: https://tidyr.tidyverse.org/index.html\n[2]: https://dplyr.tidyverse.org/index.html\n[3]: https://github.com/pwwang/pipda\n[4]: https://tibble.tidyverse.org/index.html\n[5]: https://pwwang.github.io/datar/\n[6]: https://img.shields.io/pypi/v/datar?style=flat-square\n[7]: https://pypi.org/project/datar/\n[8]: https://img.shields.io/github/v/tag/pwwang/datar?style=flat-square\n[9]: https://github.com/pwwang/datar\n[10]: https://img.shields.io/github/workflow/status/pwwang/datar/Build%20and%20Deploy?style=flat-square\n[11]: https://img.shields.io/github/workflow/status/pwwang/datar/Build%20Docs?label=Docs&style=flat-square\n[12]: https://img.shields.io/codacy/grade/3d9bdff4d7a34bdfb9cd9e254184cb35?style=flat-square\n[13]: https://app.codacy.com/gh/pwwang/datar\n[14]: https://img.shields.io/codacy/coverage/3d9bdff4d7a34bdfb9cd9e254184cb35?style=flat-square\n[15]: https://pwwang.github.io/datar/reference-maps/ALL/\n[16]: https://pwwang.github.io/datar/notebooks/across/\n[17]: https://pwwang.github.io/datar/api/datar/\n[18]: https://pwwang.github.io/datar-blog\n[19]: https://github.com/pwwang/datar-cli\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pwwang/datar',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



"""Pivot data from long to wide"""

from typing import List, Any, Union, Callable, Mapping

import pandas
from pandas import DataFrame, Index
from pipda import register_verb

from ..core.contexts import Context
from ..core.types import StringOrIter, is_scalar
from ..core.utils import vars_select, reconstruct_tibble
from ..core.exceptions import ColumnNotExistingError

from ..base import NA, identity
from ..base.na import NA_integer_

ROWID_COLUMN = "_PIVOT_ROWID_"


@register_verb(DataFrame, context=Context.SELECT)
def pivot_wider(
    _data: DataFrame,
    id_cols: StringOrIter = None,
    names_from: StringOrIter = "name",
    names_prefix: str = "",
    names_sep: str = "_",
    names_glue: str = None,
    names_sort: bool = False,
    # names_repair: str = "check_unique", # todo
    values_from: StringOrIter = "value",
    values_fill: Any = None,
    values_fn: Union[Callable, Mapping[str, Callable]] = identity,
    base0_: bool = None,
) -> DataFrame:
    """ "widens" data, increasing the number of columns and decreasing
    the number of rows.

    Args:
        _data: A data frame to pivot.
        id_cols: A set of columns that uniquely identifies each observation.
            Defaults to all columns in data except for the columns specified
            in names_from and values_from.
        names_from: and
        values_from: A pair of arguments describing which column
            (or columns) to get the name of the output column (names_from),
            and which column (or columns) to get the cell values from
            (values_from).
        names_prefix: String added to the start of every variable name.
        names_sep: If names_from or values_from contains multiple variables,
            this will be used to join their values together into a single
            string to use as a column name.
        names_glue: Instead of names_sep and names_prefix, you can supply
            a glue specification that uses the names_from columns
            (and special _value) to create custom column names.
        names_sort: Should the column names be sorted? If FALSE, the default,
            column names are ordered by first appearance.
        names_repair: todo
        values_fill: Optionally, a (scalar) value that specifies what
            each value should be filled in with when missing.
        values_fn: Optionally, a function applied to the value in each cell
            in the output. You will typically use this when the combination
            of `id_cols` and value column does not uniquely identify
            an observation.
            This can be a dict you want to apply different aggregations to
            different value columns.
            If not specified, will be `numpy.mean`
        base0_: Whether `id_cols`, `names_from` and `values_from`
            are 0-based if given by indexes.
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        The pivoted dataframe.
    """
    if is_scalar(names_from):
        names_from = [names_from]  # type: ignore
    if is_scalar(values_from):
        values_from = [values_from]  # type: ignore
    if id_cols is not None and is_scalar(id_cols):
        id_cols = [id_cols]  # type: ignore

    if id_cols is None:
        all_cols = _data.columns
        names_from = all_cols[vars_select(all_cols, names_from, base0=base0_)]
        # values_from could be a df-column
        new_values_from = []
        for value_from in values_from:
            if isinstance(value_from, str) and value_from not in all_cols:
                df_cols = [
                    col for col in all_cols if col.startswith(f"{value_from}$")
                ]
                if not df_cols:
                    raise ColumnNotExistingError(value_from)
                new_values_from.extend(df_cols)
            else:
                new_values_from.append(value_from)
        values_from = all_cols[
            vars_select(all_cols, *new_values_from, base0=base0_)
        ]
        id_cols = all_cols.difference(names_from).difference(values_from)

    # build multiindex pivot table
    id_cols = list(id_cols)
    names_from = list(names_from)
    values_from = list(values_from)

    # DF:
    #    id  x  y  a  b
    # 0  10  X  1  1  1
    # 1  20  Y  2  2  2
    #
    # to:
    #    id    a         b
    # x        X    Y    X    Y
    # y        1    2    1    2
    # 0  10  1.0  NaN  1.0  NaN
    # 1  20  NaN  2.0  NaN  2.0
    #
    # with:
    # id_cols = ['id']
    # names_from = ['x', 'y']
    # values_from = ['a', 'b']
    #
    # expected:
    #     id a_X_1 a_Y_2 b_X_1 b_Y_2
    # 0   10 1     NaN   1     NaN
    # 1   20 NaN   2     NaN   2
    if len(id_cols) == 0 and len(values_from) > 1:
        # need to add it to turn names_to to columns
        ret = _data.assign(**{ROWID_COLUMN: 0})
        id_cols = [ROWID_COLUMN]
    else:
        ret = _data

    # hold NAs in values_from columns, so that they won't be filled
    # by values_fill
    for col in values_from:
        ret[col].fillna(NA_integer_, inplace=True)

    ret = pandas.pivot_table(
        ret,
        index=id_cols,
        columns=names_from,
        fill_value=values_fill,
        values=values_from[0] if len(values_from) == 1 else values_from,
        aggfunc=values_fn,
    )

    ret.columns = _flatten_column_names(
        ret.columns, names_prefix, names_sep, names_glue
    )

    if len(id_cols) > 0:
        ret.reset_index(inplace=True)

    if ROWID_COLUMN in ret:
        ret.drop(columns=[ROWID_COLUMN], inplace=True)

    ret.reset_index(drop=True, inplace=True)
    # Get the original NAs back
    for col in ret.columns.difference(id_cols):
        ret[col].replace({NA_integer_: NA}, inplace=True)

    if names_sort:
        ret = ret.loc[:, sorted(ret.columns)]

    return reconstruct_tibble(_data, ret)


def _flatten_column_names(
    names: Index, names_prefix: str, names_sep: str, names_glue: str
) -> List[str]:
    """Flatten the hierachical column names:

    For example,
        >>> MultiIndex([('id',  '', ''),
        >>>    ( 'a', 'X',  1),
        >>>    ( 'a', 'Y',  2),
        >>>    ( 'b', 'X',  1),
        >>>    ( 'b', 'Y',  2)],
        >>>     names=[None, 'x', 'y'])
    To
        >>> ['X1_a', 'Y2_a', 'X1_b', 'Y2_b']
    with `names_glue={x}{y}_{_value}`
    """
    lvlnames = ["_value" if level is None else level for level in names.names]
    out = []
    for cols in names:
        if is_scalar(cols):
            cols = [cols]
            # out.append(f'{names_prefix}{cols}')
            # continue
        # if len(cols) == 1:
        #     out.append(f'{names_prefix}{cols[0]}')
        #     continue

        cols = dict(zip(lvlnames, (str(col) for col in cols)))
        # in case of ('id', '', '')
        # if all(name == '' for key, name in cols.items() if key != '_value'):
        #     out.append(f'{names_prefix}{cols["_value"]}')
        # in case of values_from is a dataframe column
        # ('d$a', 'X', '1')
        if "$" in cols.get("_value", ""):
            prefix = names_prefix + names_sep.join(
                col for name, col in cols.items() if name != "_value"
            )
            out.append(f'{prefix}${cols["_value"].split("$", 1)[1]}')
        elif not names_glue:
            out.append(f"{names_prefix}{names_sep.join(cols.values())}")
        else:
            if "_value" in cols:
                cols[".value"] = cols["_value"]
            out.append(names_glue.format(**cols))

    return out



"""Arrange rows by column values

See source https://github.com/tidyverse/dplyr/blob/master/R/arrange.R
"""
from typing import Any, Iterable, Mapping, Tuple
from pandas import DataFrame
from pipda import register_verb
from pipda.symbolic import DirectRefAttr, DirectRefItem
from pipda.function import FastEvalFunction
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.utils import check_column_uniqueness, reconstruct_tibble
from ..base import union
from .group_data import group_vars
from .group_by import ungroup
from .mutate import mutate
from .desc import desc


@register_verb(DataFrame, context=Context.PENDING)
def arrange(
    _data: DataFrame, *args: Any, _by_group: bool = False, **kwargs: Any
) -> DataFrame:
    """orders the rows of a data frame by the values of selected columns.

    The original API:
    https://dplyr.tidyverse.org/reference/arrange.html

    Args:
        _data: A data frame
        *series: Variables, or functions of variables.
            Use desc() to sort a variable in descending order.
        _by_group: If TRUE, will sort first by grouping variable.
            Applies to grouped data frames only.
        **kwargs: Name-value pairs that apply with mutate

    Returns:
        An object of the same type as _data.
        The output has the following properties:
            All rows appear in the output, but (usually) in a different place.
            Columns are not modified.
            Groups are not modified.
            Data frame attributes are preserved.
    """
    if not args and not kwargs:
        return _data

    check_column_uniqueness(
        _data, "Cannot arrange a data frame with duplicate names"
    )

    # See if we don't need to mutate
    # If all series are the ones from the _data itself
    if not kwargs:
        by = _series_cols(args, _data.columns)
        if by is not None:
            if _by_group:
                gvars = group_vars(_data, __calling_env=CallingEnvs.REGULAR)
                gby = dict(zip(gvars, [True] * len(gvars)))
                gby.update(by)
                by = gby

            out = _data.sort_values(
                list(by), ascending=list(by.values())
            ).reset_index(drop=True)
            return reconstruct_tibble(_data, out, keep_rowwise=True)

    if not _by_group:
        sorting_df = mutate(
            ungroup(_data, __calling_env=CallingEnvs.REGULAR),
            *args,
            **kwargs,
            _keep="none",
            __calling_env=CallingEnvs.REGULAR,
        )
        sorting_df = sorting_df.sort_values(by=sorting_df.columns.tolist())
    else:
        gvars = group_vars(_data, __calling_env=CallingEnvs.REGULAR)
        sorting_df = ungroup(
            mutate(
                _data,
                *args,
                **kwargs,
                _keep="none",
                __calling_env=CallingEnvs.REGULAR,
            ),
            __calling_env=CallingEnvs.REGULAR,
        )
        by = union(gvars, sorting_df.columns, __calling_env=CallingEnvs.REGULAR)
        sorting_df = sorting_df.sort_values(by=by)

    out = _data.loc[sorting_df.index, :].reset_index(drop=True)

    return reconstruct_tibble(_data, out, keep_rowwise=True)


def _series_col(arg: Any, columns: Iterable[str]) -> str:
    """Turn a single arg into name and desc if possible"""
    if (
        isinstance(arg, (DirectRefAttr, DirectRefItem))
        and arg._pipda_ref in columns
    ):
        return arg._pipda_ref

    return None


def _series_cols(
    args: Tuple,
    columns: Iterable[str],
) -> Mapping[str, bool]:
    """Check if one of the args is a series column or columns in original df"""
    out = {}
    for arg in args:
        sercol = _series_col(arg, columns)
        if sercol:
            out[sercol] = True

        elif (
            isinstance(arg, FastEvalFunction)
            and arg._pipda_func is desc.__origfunc__
        ):
            for col in arg._pipda_args:
                sercol = _series_col(col, columns)
                if sercol is None:
                    return None
                out[sercol] = False

        else:
            return None

    return out



"""Arithmetic or math functions"""

from typing import Any, Callable, Iterable, Union

import numpy
from pandas import DataFrame, Series
from pipda import register_func, register_verb
from pipda.utils import CallingEnvs

from ..core.collections import Collection
from ..core.contexts import Context
from ..core.types import (
    FloatOrIter,
    NumericOrIter,
    NumericType,
    is_not_null,
    is_scalar,
)
from ..core.utils import Array, length_of, recycle_value, register_numpy_func_x

# cor?, range, summary, iqr


def _register_arithmetic_agg(
    name: str, np_name: str, doc: str = ""
) -> Callable:
    """Register an arithmetic function"""

    def _arithmetric(x: Iterable, na_rm: bool = False) -> Iterable:
        """Arithmetric function"""
        # na_rm not working for numpy functions
        # with x is a Series object
        if isinstance(x, Series):
            return getattr(x, np_name)(skipna=na_rm)

        fun_name = f"nan{np_name}" if na_rm else np_name
        return getattr(numpy, fun_name)(x)

    _arithmetric.__name__ = name
    _arithmetric.__doc__ = doc
    return register_func(None, context=Context.EVAL, func=_arithmetric)


sum = _register_arithmetic_agg(
    "sum",
    "sum",
    doc="""Sum of the input.

    Args:
        x: The input
        na_rm: Exclude the NAs

    Returns:
        The sum of the input
    """,
)

prod = _register_arithmetic_agg(
    "prod",
    "prod",
    doc="""Product of the input.

    Args:
        x: The input
        na_rm: Exclude the NAs

    Returns:
        The product of the input
    """,
)

mean = _register_arithmetic_agg(
    "mean",
    "mean",
    doc="""Mean of the input.

    Args:
        x: The input
        na_rm: Exclude the NAs

    Returns:
        The mean of the input
    """,
)

median = _register_arithmetic_agg(
    "median",
    "median",
    doc="""Median of the input.

    Args:
        x: The input
        na_rm: Exclude the NAs

    Returns:
        The median of the input
    """,
)


@register_func(None, context=Context.EVAL)
def min(*x, na_rm: bool = False) -> Any:
    """Min of the input.

    Args:
        x: The input
        na_rm: Exclude the NAs

    Returns:
        The min of the input
    """
    fun = numpy.nanmin if na_rm else numpy.min
    x = Collection(*x)  # flatten
    return fun(x)


@register_func(None, context=Context.EVAL)
def max(*x, na_rm: bool = False) -> Any:
    """Max of the input.

    Args:
        x: The input
        na_rm: Exclude the NAs

    Returns:
        The max of the input
    """
    fun = numpy.nanmax if na_rm else numpy.max
    x = Collection(*x)  # flatten
    return fun(x)


@register_func(None, context=Context.EVAL)
def var(x: Any, na_rm: bool = False, ddof: int = 1):
    """Variance of the input.

    Args:
        x: The input
        na_rm: Exclude the NAs
        ddof: Delta Degrees of Freedom

    Returns:
        The variance of the input
    """
    fun = numpy.nanvar if na_rm else numpy.var
    return fun(x, ddof=ddof)


@register_func(None, context=Context.EVAL)
def pmin(*x: NumericType, na_rm: bool = False) -> Iterable[float]:
    """Get the min value rowwisely

    Args:
        *x: The iterables. Elements will be recycled to the max length
        na_rm: Whether ignore the NAs

    Returns:
        The rowwise min of `*x`
    """
    maxlen = max(map(length_of, x))
    x = (recycle_value(elem, maxlen) for elem in x)
    return Array([min(elem, na_rm=na_rm) for elem in zip(*x)])


@register_func(None, context=Context.EVAL)
def pmax(*x: Iterable, na_rm: bool = False) -> Iterable[float]:
    """Get the max value rowwisely
    Args:
        *x: The iterables. Elements will be recycled to the max length
        na_rm: Whether ignore the NAs

    Returns:
        The rowwise max of `*x`
    """
    maxlen = max(map(length_of, x))
    x = (recycle_value(elem, maxlen) for elem in x)
    return Array([max(elem, na_rm=na_rm) for elem in zip(*x)])


@register_func(None, context=Context.EVAL)
def round(x: NumericOrIter, ndigits: int = 0) -> NumericOrIter:
    """Rounding a number"""
    # recycle ndigits?
    return numpy.round(x, ndigits)


sqrt = register_numpy_func_x(
    "sqrt",
    "sqrt",
    doc="""Get the square root of a number/numbers

    Args:
        x: The input

    Returns:
        The square root of the input
    """,
)

abs = register_numpy_func_x(
    "abs",
    "abs",
    doc="""Get the absolute value of a number/numbers

    Args:
        x: The input

    Returns:
        The absolute values of the input
    """,
)

sign = register_numpy_func_x(
    "sign",
    "sign",
    doc="""Get the signs of the corresponding elements of x

    Args:
        x: The input

    Returns:
        The signs of the corresponding elements of x
    """,
)

trunc = register_numpy_func_x(
    "trunc",
    "trunc",
    doc="""Get the integers truncated for each element in x

    Args:
        x: The input

    Returns:
        The ingeters of elements in x being truncated
        Note the dtype is still float.
    """,
)

ceiling = register_numpy_func_x(
    "ceiling",
    "ceil",
    doc="""Get the ceiling integer of a number/numbers

    Args:
        x: The input

    Returns:
        The ceiling integer of the input
    """,
)

floor = register_numpy_func_x(
    "floor",
    "floor",
    doc="""Get the floor integer of a number/numbers

    Args:
        x: The input

    Returns:
        The floor integer of the input
    """,
)


@register_func(None, context=Context.EVAL)
def signif(x: NumericOrIter, digits: int = 6) -> NumericOrIter:
    """Rounds the values in its first argument to the specified number of
    significant digits

    Args:
        x: A numeric vector or scalar
        digits: integer indicating the number of significant digits to be used

    Returns:
        The rounded values for each element in x
    """
    # todo complex?
    return numpy.fromiter(
        (round(elem, digits - int(ceiling(log10(abs(elem))))) for elem in x),
        dtype=float,
    )


@register_verb(DataFrame, context=Context.EVAL)
def cov(x: DataFrame, y: Iterable = None, ddof: int = 1) -> DataFrame:
    """Compute pairwise covariance of dataframe columns,
    or between two variables
    """
    # TODO: support na_rm, use, method. see `?cov` in R
    return x.cov(ddof=ddof)


@cov.register((numpy.ndarray, Series, list, tuple), context=Context.EVAL)
def _(x: Iterable, y: Iterable, ddof: int = 1) -> DataFrame:
    """Compute covariance for two iterables"""
    # ddof: numpy v1.5+
    return numpy.cov(x, y, ddof=ddof)[0][1]


@register_verb(DataFrame, context=Context.EVAL)
def scale(
    x: DataFrame,
    center: Union[bool, NumericOrIter] = True,

    scale: Union[bool, NumericOrIter] = True,
) -> DataFrame:
    """Scaling and Centering of a numeric data frame

    See Details in `?scale` in `R`

    Args:
        x: The numeric data frame to scale
        center: either a logical value or numeric-alike vector of length
            equal to the number of columns of `x`
        scale: either a logical value or a numeric-alike vector of length
            equal to the number of columns of `x`.

    Returns:
        The centered, scaled data frame
    """
    # center
    ncols = x.shape[1]
    center_is_true = center is True
    out_attrs = {}

    if center is True:
        center = col_means(x, __calling_env=CallingEnvs.REGULAR)

    elif center is not False:
        if is_scalar(center):
            center = [center]  # type: ignore
        if len(center) != ncols:
            raise ValueError(
                f"length of `center` ({len(center)}) must equal "
                f"the number of columns of `x` ({ncols})"
            )

    if center is not False:
        x = x - center
        out_attrs["scaled:center"] = Array(center)

    # scale
    if scale is True:

        def _rms(col: Series) -> Series:
            nonnas = col[is_not_null(col)] ** 2
            return sqrt(nonnas.sum() / (len(nonnas) - 1))

        scale = (
            col_sds(x, __calling_env=CallingEnvs.REGULAR)
            if center_is_true
            else x.agg(_rms)
        )

    elif scale is not False:
        if is_scalar(scale):
            scale = [scale]  # type: ignore
        if len(scale) != ncols:
            raise ValueError(
                f"length of `scale` ({len(center)}) must equal "
                f"the number of columns of `x` ({ncols})"
            )

    if scale is not False:
        x = x / scale
        out_attrs["scaled:scale"] = Array(scale)

    if center is False and scale is False:
        x = x.copy()

    x.attrs.update(out_attrs)
    return x


# being able to refer it inside the function
# as scale also used as an argument
_scale = scale


@scale.register(Series)
def _(
    x: Series,
    center: Union[bool, NumericOrIter] = True,

    scale: Union[bool, NumericOrIter] = True,
) -> DataFrame:
    """Scaling on series"""
    return _scale(x.to_frame(), center, scale)


@scale.register((numpy.ndarray, list, tuple))
def _(
    x: Iterable,
    center: Union[bool, NumericOrIter] = True,

    scale: Union[bool, NumericOrIter] = True,
) -> DataFrame:
    """Scaling on iterables"""
    return _scale(Series(x, name="scaled"), center, scale)


@register_verb(DataFrame)
def col_sums(
    x: DataFrame,
    na_rm: bool = False,
    # dims: int = 1,
    # weights = None,
    # freq = None,
    # n = None
) -> Iterable[NumericType]:
    """Calculate sum of a data frame by column

    Args:
        x: The data frame
        na_rm: Specifies how to handle missing values in `x`.

    Returns:
        The sums by column.
    """
    return x.agg(sum, na_rm=na_rm)


@register_verb(DataFrame)
def row_sums(
    x: DataFrame,
    na_rm: bool = False,
    # dims: int = 1,
    # weights = None,
    # freq = None,
    # n = None
) -> Iterable[NumericType]:
    """Calculate sum of a data frame by row

    Args:
        x: The data frame
        na_rm: Specifies how to handle missing values in `x`.

    Returns:
        The sums by row.
    """
    return x.agg(sum, axis=1, na_rm=na_rm)


@register_verb(DataFrame)
def col_means(
    x: DataFrame,
    na_rm: bool = False,
    # dims: int = 1,
    # weights = None,
    # freq = None,
    # n = None
) -> Iterable[NumericType]:
    """Calculate mean of a data frame by column

    Args:
        x: The data frame
        na_rm: Specifies how to handle missing values in `x`.

    Returns:
        The means by column.
    """
    return x.agg(mean, na_rm=na_rm)


@register_verb(DataFrame)
def row_means(
    x: DataFrame,
    na_rm: bool = False,
    # dims: int = 1,
    # weights = None,
    # freq = None,
    # n = None
) -> Iterable[NumericType]:
    """Calculate mean of a data frame by row

    Args:
        x: The data frame
        na_rm: Specifies how to handle missing values in `x`.

    Returns:
        The means by row.
    """
    return x.agg(mean, axis=1, na_rm=na_rm)


@register_verb(DataFrame)
def col_sds(
    x: DataFrame,
    na_rm: bool = False,
    # dims: int = 1,
    # weights = None,
    # freq = None,
    # n = None
) -> Iterable[NumericType]:
    """Calculate stdev of a data frame by column

    Args:
        x: The data frame
        na_rm: Specifies how to handle missing values in `x`.

    Returns:
        The stdevs by column.
    """
    from ..stats import sd

    return x.agg(sd, na_rm=na_rm)


@register_verb(DataFrame)
def row_sds(
    x: DataFrame,
    na_rm: bool = False,
    # dims: int = 1,
    # weights = None,
    # freq = None,
    # n = None
) -> Iterable[NumericType]:
    """Calculate stdev of a data frame by row

    Args:
        x: The data frame
        na_rm: Specifies how to handle missing values in `x`.

    Returns:
        The stdevs by row.
    """
    from ..stats import sd

    return x.agg(sd, axis=1, na_rm=na_rm)


@register_verb(DataFrame)
def col_medians(
    x: DataFrame,
    na_rm: bool = False,
    # dims: int = 1,
    # weights = None,
    # freq = None,
    # n = None
) -> Iterable[NumericType]:
    """Calculate median of a data frame by column

    Args:
        x: The data frame
        na_rm: Specifies how to handle missing values in `x`.

    Returns:
        The medians by column.
    """
    return x.agg(median, na_rm=na_rm)


@register_verb(DataFrame)
def row_medians(
    x: DataFrame,
    na_rm: bool = False,
    # dims: int = 1,
    # weights = None,
    # freq = None,
    # n = None
) -> Iterable[NumericType]:
    """Calculate median of a data frame by row

    Args:
        x: The data frame
        na_rm: Specifies how to handle missing values in `x`.

    Returns:
        The medians by row.
    """
    return x.agg(median, axis=1, na_rm=na_rm)


@register_func(None, context=Context.EVAL)
def log(x: NumericOrIter, base: float = numpy.e) -> FloatOrIter:
    """Computes logarithms, by default natural logarithm

    Args:
        x: A numeric scalar or vector
        base: The base of the logarithm

    Returns:
        The value of the logarithm if x is scalar, otherwise element-wise
        logarithm of elements in x
    """
    if base == numpy.e:
        return numpy.log(x)

    return numpy.log(x) / numpy.log(base)


exp = register_numpy_func_x(
    "exp",
    "exp",
    doc="""Calculates the power of natural number

    Args:
        x: A numeric scalar or vector

    Returns:
        Power of natural number of element-wise power of natural number for x
    """,
)

log2 = register_numpy_func_x(
    "log2",
    "log2",
    doc="""Computes logarithms with base 2

    Args:
        x: A numeric scalar or vector

    Returns:
        The value of log2 if x is scalar, otherwise element-wise
        log2 of elements in x
    """,
)

log10 = register_numpy_func_x(
    "log10",
    "log10",
    doc="""Computes logarithms with base 10

    Args:
        x: A numeric scalar or vector

    Returns:
        The value of log10 if x is scalar, otherwise element-wise
        log10 of elements in x
    """,
)

log1p = register_numpy_func_x(
    "log1p",
    "log1p",
    doc="""Computes log(1+x)

    Args:
        x: A numeric scalar or vector

    Returns:
        The value of log(1+x) if x is scalar, otherwise element-wise
        log(1+x) of elements in x
    """,
)



"""Name repairing"""
import inspect
import re
import keyword
from typing import Callable, List, Sequence, Union, Iterable, Tuple

import numpy

from .exceptions import NameNonUniqueError
from .types import is_iterable


def _log_changed_names(changed_names: Iterable[Tuple[str, str]]) -> None:
    """Log the changed names"""
    if not changed_names:
        return
    from .utils import logger

    logger.warning("New names:")
    for orig_name, new_name in changed_names:
        logger.warning("* %r -> %r", orig_name, new_name)


def _repair_names_minimal(names: Iterable[str]) -> List[str]:
    """Minimal repairing"""
    return ["" if name in (None, numpy.nan) else str(name) for name in names]


def _repair_names_unique(
    names: Sequence[str],
    quiet: bool = False,
    sanitizer: Callable = None,
    base0_: bool = None,
) -> List[str]:
    """Make sure names are unique"""
    base = int(not base0_)
    min_names = _repair_names_minimal(names)
    neat_names = [
        re.sub(r"(?:(?<!_)_{1,2}\d+|(?<!_)__)+$", "", name)
        for name in min_names
    ]
    if callable(sanitizer):
        neat_names = [sanitizer(name) for name in neat_names]

    new_names = []
    changed_names = []
    for i, name in enumerate(neat_names):
        if neat_names.count(name) > 1 or name == "":
            name = f"{name}__{i + base}"
        if name != names[i]:
            changed_names.append((names[i], name))
        new_names.append(name)
    if not quiet:
        _log_changed_names(changed_names)
    return new_names


def _repair_names_universal(
    names: Iterable[str], quiet: bool = False, base0_: bool = None
) -> List[str]:
    """Make sure names are safely to be used as variable or attribute"""
    min_names = _repair_names_minimal(names)
    neat_names = [re.sub(r"[^\w]", "_", name) for name in min_names]
    new_names = _repair_names_unique(
        neat_names,
        quiet=True,
        sanitizer=lambda name: (
            f"_{name}"
            if keyword.iskeyword(name) or (name and name[0].isdigit())
            else name
        ),
        base0_=base0_,
    )
    if not quiet:
        changed_names = [
            (orig_name, new_name)
            for orig_name, new_name in zip(names, new_names)
            if orig_name != new_name
        ]
        _log_changed_names(changed_names)
    return new_names


def _repair_names_check_unique(names: Iterable[str]) -> Iterable[str]:
    """Just check the uniqueness"""
    for name in names:
        if names.count(name) > 1:
            raise NameNonUniqueError(f"Names must be unique: {name}")
        if name == "" or name is numpy.nan:
            raise NameNonUniqueError(f"Names can't be empty: {name}")
        if re.search(r"(?:(?<!_)_{2}\d+|(?<!_)__)+$", str(name)):
            raise NameNonUniqueError(
                f"Names can't be of the form `__` or `_j`: {name}"
            )
    return names


BUILTIN_REPAIR_METHODS = dict(
    minimal=_repair_names_minimal,
    unique=_repair_names_unique,
    universal=_repair_names_universal,
    check_unique=_repair_names_check_unique,
)


def repair_names(
    names: Iterable[str],
    repair: Union[str, Callable],
    base0_: bool = None,
) -> List[str]:
    """Repair names based on the method

    Args:
        names: The names to be repaired
        repair: The method to repair
            - `minimal`: Minimal names are never None or NA.
                When an element doesn't have a name, its minimal name
                is an empty string.
            - `unique`: Unique names are unique. A suffix is appended to
                duplicate names to make them unique.
            - `universal`: Universal names are unique and syntactic,
                meaning that you can safely use the names as variables without
                causing a syntax error (like `f.<name>`).
            - A function, accepts either a list of names or a single name.
                Function accepts a list of names must annotate the first
                argument with `typing.Iterable` or `typing.Sequence`.
        base0_: Whether the numeric suffix starts from 0 or not.
            If not specified, will use `datar.base.get_option('index.base.0')`.

    Examples:
        >>> repair_names([None]*3, repair="minimal")
        >>> # ["", "", ""]
        >>> repair_names(["x", NA], repair="minimal")
        >>> # ["x", ""]
        >>> repair_names(["", "x", "", "y", "x", "_2", "__"], repair="unique")
        >>> # ["__1", "x__2", "__3", "y", "x__5", "__6", "__7"]
        >>> repair_names(["", "x", NA, "x"], repair="universal")
        >>> # ["__1", "x__2", "__3", "x__4"]
        >>> repair_names(["(y)"  "_z"  ".2fa"  "False"], repair="universal")
        >>> # ["_y_", "_z", "_2fa", "_False"]

    Returns:
        The repaired names

    Raises:
        ValueError: when repair is not a string or callable
        NameNonUniqueError: when check_unique fails
    """
    from .utils import get_option

    base0_ = get_option("index.base.0", base0_)
    if isinstance(repair, str):
        repair = BUILTIN_REPAIR_METHODS[repair]  # type: ignore
    elif is_iterable(repair) and all(isinstance(elem, str) for elem in repair):
        return repair  # type: ignore
    elif not callable(repair):
        raise ValueError("Expect a function for name repairing.")

    parameters = inspect.signature(repair).parameters  # type: ignore
    annotation = list(parameters.values())[0].annotation
    if annotation is inspect._empty or annotation._name not in (
        "Iterable",
        "Sequence",
    ):  # scalar input
        return [
            repair(name, base0_=base0_)  # type: ignore[operator]
            if "base0_" in parameters
            else repair(name)  # type: ignore[operator]
            for name in names
        ]

    names = list(names)
    return (
        repair(names, base0_=base0_)  # type: ignore[operator]
        if "base0_" in parameters
        else repair(names)  # type: ignore[operator]
    )



"""Import all constants, verbs and functions"""
from . import f
from .base import _no_warn as _  # don't override from datar.all import _no_warn
from .base import _builtin_names as _base_builtin_names
from .base import *
from .base import _warn as _
from .datar import *
from .dplyr import _no_warn as _
from .dplyr import _builtin_names as _dplyr_builtin_names
from .dplyr import *
from .dplyr import _warn as _
from .stats import *
from .tibble import *
from .tidyr import *
from .utils import *
from .forcats import *

_builtin_names = _base_builtin_names.copy()
_builtin_names.update(_dplyr_builtin_names)
# builtin names included
__all__ = [var_ for var_ in locals() if not var_.startswith("_")]

for name in _builtin_names:
    # let __getattr__ handles the builtins, otherwise
    # from datar.all import filter
    # will not warn
    del locals()[name]

from .core.warn_builtin_names import (
    warn_builtin_names as _warn_builtin_names,
)

__getattr__ = _warn_builtin_names(**_builtin_names)



"""Uncount a data frame"""

from typing import Any, Iterable

import numpy
from pandas import DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.types import IntOrIter, is_scalar
from ..core.utils import get_option, reconstruct_tibble

from ..dplyr import group_by, mutate, row_number, ungroup

INDEX_COLUMN = "_UNCOUND_INDEX_"


@register_verb(DataFrame, context=Context.EVAL)
def uncount(
    data: DataFrame,
    weights: IntOrIter,
    _remove: bool = True,
    _id: str = None,
    base0_: bool = None,
) -> DataFrame:
    """Duplicating rows according to a weighting variable

    Args:
        data: A data frame
        weights: A vector of weights. Evaluated in the context of data
        _remove: If TRUE, and weights is the name of a column in data,
            then this column is removed.
        _id: Supply a string to create a new variable which gives a
            unique identifier for each created row (0-based).
        base0_: Whether the generated `_id` columns are 0-based.
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        dataframe with rows repeated.
    """
    if is_scalar(weights):
        weights = [weights] * data.shape[0]  # type: ignore

    _check_weights(weights)

    indexes = [
        idx for i, idx in enumerate(data.index) for _ in range(int(weights[i]))
    ]

    all_columns = data.columns
    weight_name = getattr(weights, "name", None)
    if weight_name in all_columns and weights is data[weight_name]:
        rest_columns = all_columns.difference([weight_name])
    else:
        rest_columns = all_columns

    out = data.loc[indexes, rest_columns] if _remove else data.loc[indexes, :]
    # need the indexes to get the right id column
    out = out.assign(**{INDEX_COLUMN: indexes})
    out.reset_index(drop=True, inplace=True)

    if _id:
        base = int(not get_option("index.base.0", base0_))

        out = ungroup(
            mutate(
                group_by(out, INDEX_COLUMN, __calling_env=CallingEnvs.REGULAR),
                **{_id: row_number() + base - 1},
                __calling_env=CallingEnvs.REGULAR,
            ),
            __calling_env=CallingEnvs.REGULAR,
        )

    out.drop(columns=[INDEX_COLUMN], inplace=True)
    return reconstruct_tibble(data, out)


def _check_weights(weights: Iterable[Any]) -> None:
    """Check if uncounting weights are valid"""
    for weight in weights:
        if not isinstance(weight, (int, float, numpy.number)):
            raise ValueError("`weights` must evaluate to numerics.")
        if weight < 0:
            raise ValueError("All elements in `weights` must be >= 0.")



"""Helper functions for ordering window function output

https://github.com/tidyverse/dplyr/blob/master/R/order-by.R
"""
from typing import Callable, Iterable, Any, Sequence

from pandas import Series
from pipda import register_func

from ..core.contexts import Context
from ..core.types import is_scalar


@register_func(None, context=Context.EVAL)
def order_by(
    order: Sequence[Any],
    data: Sequence[Any],
) -> Series:
    """Order the data by the given order

    Note:
        This behaves differently than `dplyr::order_by`, since we are unable
        to maintain the `data` argument to stay uncalled here.

        To control the order of the argument of a call, try `with_order`.

    Examples:
        >>> order_by(range(3,0), [5,4,3])
        >>> # [3,4,5]
        >>> order_by([5,4,3,2,1], cumsum([1,2,3,4,5]))
        >>> # 15,10,6,3,1
        >>> # instead of 15,14,12,9,5 from dplyr::order_by

    Args:
        `order`: An iterable to control the data order
        `data`: The data to be ordered

    Returns:
        The odered series. Note the original class of data will be lost.
    """
    if is_scalar(order):
        order = [order]
    if is_scalar(data):
        data = [data]

    order = Series(order) if len(order) > 1 else Series(order, dtype=object)
    order = order.sort_values()
    out = Series(data) if len(data) > 1 else Series(data, dtype=object)
    return out[order.index]


@register_func(None, context=Context.EVAL)
def with_order(
    order: Iterable[Any],
    func: Callable,
    x: Iterable[Any],
    *args: Any,
    **kwargs: Any,
) -> Series:
    """Control argument and result of a window function

    Examples:
        >>> with_order([5,4,3,2,1], cumsum, [1,2,3,4,5])
        >>> # 15, 14, 12, 9, 5

    Args:
        order: An iterable to order the arugment and result
        func: The window function
        x: The first arugment for the function
        *args: and
        **kwargs: Other arugments for the function

    Returns:
        A Seroes of ordered result
    """
    x = order_by(order, x)
    # expecting func to return an iterable
    out = func(x, *args, **kwargs)
    if isinstance(out, Series):
        # drop the index
        out = out.reset_index(drop=True)
    return order_by(order, out)



"""Logical/Boolean functions"""
from typing import Any

import numpy
from pandas.core.dtypes.common import is_bool_dtype
from pipda import register_func

from ..core.types import BoolOrIter, is_scalar
from ..core.contexts import Context

from .testing import _register_type_testing
from .casting import _as_type

TRUE = True
FALSE = False


@register_func(None, context=Context.EVAL)
def as_logical(x: Any) -> BoolOrIter:
    """Convert an object or elements of an iterable into bool

    Args:
        x: The object

    Returns:
        When x is an array or a series, return x.astype(bool).
        When x is iterable, convert elements of it into bools
        Otherwise, convert x to bool.
    """
    return _as_type(x, bool)


as_bool = as_logical


is_logical = _register_type_testing(
    "is_logical",
    scalar_types=(bool, numpy.bool_),
    dtype_checker=is_bool_dtype,
    doc="""Test if a value is booleans

Args:
    x: The value to be checked

Returns:
    True if the value is an boolean or booleans; False otherwise.
""",
)

is_bool = is_logical


@register_func(None, context=Context.EVAL)
def is_true(x: Any) -> bool:
    """Check if a value is a scalar True, like `isTRUE()` in `R`.

    If the value is non-scalar, will return False
    This does not require exact True, but a value that passes
    `if` test in python

    Args:
        x: The value to test

    Returns:
        True if x is True otherwise False
    """
    if not is_scalar(x):
        return False
    return bool(x)


@register_func(None, context=Context.EVAL)
def is_false(x: Any) -> bool:
    """Check if a value is a scalar False, like `isFALSE()` in `R`.

    If the value is non-scalar, will return False
    This does not require exact False, but a value that fails
    `if` test in python

    Args:
        x: The value to test

    Returns:
        False if x is False otherwise True
    """
    if not is_scalar(x):
        return False
    return not bool(x)



"""Apply a function (or functions) across multiple columns

See source https://github.com/tidyverse/dplyr/blob/master/R/across.R
"""
from abc import ABC, abstractmethod
from typing import Any, Callable, Iterable, List, Mapping, Tuple, Union

import numpy
from pandas import DataFrame, Series
from pipda import register_func, evaluate_expr
from pipda.function import Function
from pipda.utils import functype, CallingEnvs
from pipda.context import ContextBase

from ..core.utils import (
    df_setitem,
    length_of,
    recycle_df,
    to_df,
    vars_select,
    get_option,
)
from ..core.middlewares import CurColumn
from ..core.contexts import Context
from .tidyselect import everything


class Across:
    """Across object"""

    def __init__(
        self,
        data: DataFrame,
        cols: Iterable[str] = None,
        fns: Union[Callable, Iterable[Callable], Mapping[str, Callable]] = None,
        names: str = None,
        base0: bool = None,
        args: Tuple[Any] = None,
        kwargs: Mapping[str, Any] = None,
    ) -> None:
        cols = everything(data) if cols is None else cols
        if not isinstance(cols, (list, tuple)):
            cols = [cols]
        cols = data.columns[vars_select(data.columns, cols, base0=base0)]
        base0 = get_option("index.base.0", base0)

        fns_list = []  # type: List[str, Union[int, Callable]]
        if callable(fns):
            fns_list.append({"fn": fns})
        elif isinstance(fns, (list, tuple)):
            fns_list.extend(
                {"fn": fn, "_fn": i + int(not base0), "_fn1": i + 1, "_fn0": i}
                for i, fn in enumerate(fns)
            )
        elif isinstance(fns, dict):
            fns_list.extend(
                {"fn": value, "_fn": key} for key, value in fns.items()
            )
        elif fns is not None:
            raise ValueError(
                "Argument `_fns` of across must be None, a function, "
                "a formula, or a dict of functions."
            )

        self.data = data
        self.cols = cols
        self.fns = fns_list
        self.names = names
        self.args = args or ()
        self.kwargs = kwargs or {}

    def evaluate(
        self, context: Union[Context, ContextBase] = None
    ) -> DataFrame:
        """Evaluate object with context"""
        if isinstance(context, Context):
            context = context.value

        if not self.fns:
            self.fns = [{"fn": lambda x: x}]
        ret = None
        for column in self.cols:
            for fn_info in self.fns:
                render_data = fn_info.copy()
                render_data["_col"] = column
                fn = render_data.pop("fn")
                name_format = self.names
                if not name_format:
                    name_format = (
                        "{_col}_{_fn}" if "_fn" in render_data else "{_col}"
                    )

                name = name_format.format(**render_data)
                args = CurColumn.replace_args(self.args, column)
                kwargs = CurColumn.replace_kwargs(self.kwargs, column)
                if functype(fn) == "plain":
                    value = fn(
                        self.data[column],
                        *evaluate_expr(args, self.data, context),
                        **evaluate_expr(kwargs, self.data, context),
                    )
                else:
                    # use fn's own context
                    value = fn(
                        self.data[column],
                        *args,
                        **kwargs,
                        __calling_env=CallingEnvs.PIPING,
                    )
                    # fast evaluation tried, if failed:
                    # will this happen? it fails when first argument
                    # cannot be evaluated
                    if isinstance(value, Function):  # pragma: no cover
                        value = value._pipda_eval(self.data, context)

                if ret is None:
                    ret = to_df(value, name)
                elif length_of(ret) == 1:
                    ret, value = recycle_df(ret, value)
                    ret = df_setitem(ret, name, value)
                else:
                    ret = df_setitem(ret, name, value)
        return DataFrame() if ret is None else ret


class IfCross(Across, ABC):
    """Base class for IfAny and IfAll"""

    @staticmethod
    @abstractmethod
    def aggregate(values: Series) -> bool:
        """How to aggregation by rows"""

    def evaluate(self, context: ContextBase = None) -> DataFrame:
        """Evaluate the object with context"""
        # Fill NA first and then do and/or
        # Since NA | True -> False for pandas
        return (
            super()
            .evaluate(context)
            .apply(self.__class__.aggregate, axis=1)
            .astype(bool)
        )


class IfAny(IfCross):
    """For calls from dplyr's if_any"""

    @staticmethod
    def aggregate(values: Series) -> bool:
        """How to aggregation by rows"""
        return values.fillna(False).astype(bool).any()


class IfAll(IfCross):
    """For calls from dplyr's if_all"""

    @staticmethod
    def aggregate(values: Series) -> bool:
        """How to aggregation by rows"""
        return values.fillna(False).astype(bool).all()


@register_func(
    context=Context.PENDING, verb_arg_only=True, summarise_prefers_input=True
)
def across(
    _data: DataFrame,
    *args: Any,
    _names: str = None,
    _fn_context: Union[Context, ContextBase] = Context.EVAL,
    base0_: bool = None,
    **kwargs: Any,
) -> DataFrame:
    """Apply the same transformation to multiple columns

    The original API:
    https://dplyr.tidyverse.org/reference/across.html

    Args:
        _data: The dataframe.
        *args: If given, the first 2 elements should be columns and functions
            apply to each of the selected columns. The rest of them will be
            the arguments for the functions.
        _names: A glue specification that describes how to name
            the output columns. This can use `{_col}` to stand for the
            selected column name, and `{_fn}` to stand for the name of
            the function being applied.
            The default (None) is equivalent to `{_col}` for the
            single function case and `{_col}_{_fn}` for the case where
            a list is used for _fns. In such a case, `{_fn}` is 0-based.
            To use 1-based index, use `{_fn1}`
        base0_: Indicating whether the columns are 0-based if selected
            by indexes. if not provided, will use
            `datar.base.get_option('index.base.0')`.
        _fn_context: Defines the context to evaluate the arguments for functions
            if they are plain functions.
            Note that registered functions will use its own context
        **kwargs: Keyword arguments for the functions

    Returns:
        A dataframe with one column for each column and each function.
    """
    if not args:
        args = (None, None)
    elif len(args) == 1:
        args = (args[0], None)
    _cols, _fns, *args = args
    _cols = evaluate_expr(_cols, _data, Context.SELECT)

    return Across(_data, _cols, _fns, _names, base0_, args, kwargs).evaluate(
        _fn_context
    )


@register_func(context=Context.SELECT, verb_arg_only=True)
def c_across(
    _data: DataFrame,
    _cols: Iterable[str] = None,
    base0_: bool = None,
) -> Series:
    """Apply the same transformation to multiple columns rowwisely

    Args:
        _data: The dataframe
        _cols: The columns
        base0_: Indicating whether the columns are 0-based if selected
            by indexes. if not provided, will use
            `datar.base.get_option('index.base.0')`.

    Returns:
        A series
    """
    if not _cols:
        _cols = everything(_data)

    _cols = vars_select(_data.columns.tolist(), _cols, base0=base0_)

    series = [_data.iloc[:, col] for col in _cols]
    return numpy.concatenate(series)


@register_func(
    context=None,
    extra_contexts={"args": Context.SELECT},
    verb_arg_only=True,
    summarise_prefers_input=True,
)
def if_any(
    _data: DataFrame,
    *args: Any,
    _names: str = None,
    _context: ContextBase = None,
    base0_: bool = None,
    **kwargs: Any,
) -> Iterable[bool]:
    """Apply the same predicate function to a selection of columns and combine
    the results True if any element is True.

    See Also:
        [`across()`](datar.dplyr.across.across)
    """
    if not args:
        args = (None, None)
    elif len(args) == 1:
        args = (args[0], None)
    _cols, _fns, *args = args

    return IfAny(_data, _cols, _fns, _names, base0_, args, kwargs).evaluate(
        _context
    )


@register_func(
    context=None,
    extra_contexts={"args": Context.SELECT},
    verb_arg_only=True,
    summarise_prefers_input=True,
)
def if_all(
    _data: DataFrame,
    # _cols: Iterable[str] = None,
    # _fns: Union[Mapping[str, Callable]] = None,
    *args: Any,
    _names: str = None,
    _context: ContextBase = None,
    base0_: bool = None,
    **kwargs: Any,
) -> Iterable[bool]:
    """Apply the same predicate function to a selection of columns and combine
    the results True if all elements are True.

    See Also:
        [`across()`](datar.dplyr.across.across)
    """
    if not args:
        args = (None, None)
    elif len(args) == 1:
        args = (args[0], None)
    _cols, _fns, *args = args

    return IfAll(_data, _cols, _fns, _names, base0_, args, kwargs).evaluate(
        _context
    )



"""Complete a data frame with missing combinations of data

https://github.com/tidyverse/tidyr/blob/HEAD/R/complete.R
"""

from typing import Iterable, Mapping, Any

from pandas import DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.utils import reconstruct_tibble

from ..dplyr import full_join
from .replace_na import replace_na
from .expand import expand


@register_verb(DataFrame, context=Context.PENDING)
def complete(
    data: DataFrame,
    *args: Iterable[Any],
    fill: Mapping[str, Any] = None,
    **kwargs: Iterable[Any],
) -> DataFrame:
    """Turns implicit missing values into explicit missing values.

    Args:
        data: A data frame
        *args: columns to expand. Columns can be atomic lists.
            - To find all unique combinations of x, y and z, including
              those not present in the data, supply each variable as a
              separate argument: `expand(df, x, y, z)`.
            - To find only the combinations that occur in the data, use
              nesting: `expand(df, nesting(x, y, z))`.
            - You can combine the two forms. For example,
              `expand(df, nesting(school_id, student_id), date)` would
              produce a row for each present school-student combination
              for all possible dates.

    Returns:
        Data frame with missing values completed
    """
    full = expand(data, *args, **kwargs, __calling_env=CallingEnvs.REGULAR)
    if full.shape[0] == 0:
        return data.copy()

    full = full_join(
        full,
        data,
        by=full.columns.tolist(),
        __calling_env=CallingEnvs.REGULAR,
    )
    full = replace_na(full, fill, __calling_env=CallingEnvs.REGULAR)

    return reconstruct_tibble(data, full)



"""Recode values

https://github.com/tidyverse/dplyr/blob/master/R/recode.R
"""
from typing import Any, Iterable, Mapping

import numpy
import pandas
from pandas import Categorical, Series
from pipda import register_func
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.utils import logger, get_option, Array
from ..core.types import (
    NoneType,
    is_not_null,
    is_scalar,
    is_null,
    is_categorical,
)
from ..base import NA, unique, c, intersect
from ..base.na import NA_integer_, NA_character_


def _get_first(x: Iterable[Any]) -> Any:
    """Get first raw item from an iterable"""
    x = x[0]
    try:
        return x.item()
    except AttributeError:
        return x


def _args_to_recodings(
    *args: Any, _force_index: bool = False, **kwargs: Any
) -> Mapping[Any, Any]:
    """Convert arguments to replaceable"""
    base0_ = get_option("index.base.0")
    values = {}
    for i, arg in enumerate(args):
        if isinstance(arg, dict):
            values.update(arg)
        else:
            values[i + int(not base0_)] = arg

    values.update(kwargs)
    if _force_index:
        for key in list(values):
            if isinstance(key, str) and key.isdigit():
                values[int(key)] = values.pop(key)
    return values


def _check_length(val: numpy.ndarray, x: numpy.ndarray, name: str):
    """Check the length of the values to recode"""
    length_x = len(val)
    n = len(x)
    if length_x in (1, n):
        return

    if n == 1:
        raise ValueError(f"{name} must be length 1, not {length_x}.")
    raise ValueError(f"{name} must be length {n}, not {length_x}.")


def _check_type(val: numpy.ndarray, out_type: type, name: str):
    """Check the type of the values to recode"""
    if out_type is NoneType:
        return  # pragma: no cover

    if val.dtype is numpy.dtype(object):
        if out_type and not all(isinstance(elem, out_type) for elem in val):
            raise TypeError(
                f"{name} must be {out_type.__name__}, not {type(val[0])}."
            )
    elif out_type and not isinstance(_get_first(val), out_type):
        raise TypeError(
            f"{name} must be {out_type.__name__}, not {val.dtype.name}."
        )


def _replace_with(

    x: numpy.ndarray,
    out_type: type,
    i: numpy.ndarray,
    val: Any,
    name: str,
) -> numpy.ndarray:
    """Replace with given value at index"""
    # https://github.com/tidyverse/dplyr/blob/HEAD/R/utils-replace-with.R
    if val is None:
        return x

    if is_scalar(val):
        val = numpy.array([val])
    else:
        val = numpy.array(val)

    _check_length(val, x, name)
    if not is_null(val).any():
        _check_type(val, out_type, name)
    # check_class(val, x, name)

    i[is_null(i)] = False

    if len(val) == 1:
        x[i] = val[0]
    else:
        x[i] = val[i]

    return x


def _validate_recode_default(
    default: Any,
    x: numpy.ndarray,
    out: numpy.ndarray,
    out_type: type,
    replaced: numpy.ndarray,
) -> numpy.ndarray:
    """Validate default for recoding"""
    default = _recode_default(x, default, out_type)
    if default is None and sum(replaced & is_not_null(x)) < len(
        out[is_not_null(x)]
    ):
        logger.warning(
            "Unreplaced values treated as NA as `_x` is not compatible. "
            "Please specify replacements exhaustively or supply `_default`",
        )

    return default


def _recode_default(x: numpy.ndarray, default: Any, out_type: type) -> Any:
    """Get right default for recoding"""
    if default is None and (
        out_type is None or isinstance(_get_first(x), out_type)
    ):
        return x
    return default


def _recode_numeric(
    _x: numpy.ndarray,
    *args: Any,
    _default: Any = None,
    _missing: Any = None,
    **kwargs: Any,
) -> numpy.ndarray:
    """Recode numeric vectors"""

    values = _args_to_recodings(*args, **kwargs, _force_index=True)
    _check_args(values, _default, _missing)
    if any(not isinstance(val, int) for val in values):
        raise ValueError("All values must be unnamed (or named with integers).")

    n = len(_x)
    out = numpy.array([NA] * n, dtype=object)
    replaced = numpy.array([False] * n)
    out_type = None

    for val in values:
        if out_type is None:
            out_type = type(values[val])
        out = _replace_with(
            out, out_type, _x == val, values[val], f"Element {val}"
        )
        replaced[_x == val] = True

    _default = _validate_recode_default(_default, _x, out, out_type, replaced)
    out = _replace_with(
        out, out_type, ~replaced & is_not_null(_x), _default, "`_default`"
    )
    out = _replace_with(
        out, out_type, is_null(_x) | (_x == NA_integer_), _missing, "`_missing`"
    )
    if out_type and not is_null(out).any():
        out = out.astype(out_type)
    return out


def _recode_character(
    _x: Iterable[Any],
    *args: Any,
    _default: Any = None,
    _missing: Any = None,
    **kwargs: Any,
) -> numpy.ndarray:
    """Recode character vectors"""
    values = _args_to_recodings(*args, **kwargs)
    _check_args(values, _default, _missing)
    if not all(isinstance(val, str) for val in values):
        raise ValueError("All values must be named.")

    n = len(_x)
    out = numpy.array([NA] * n, dtype=object)
    replaced = numpy.array([False] * n)
    out_type = None

    for val in values:
        if out_type is None:
            out_type = type(values[val])
        out = _replace_with(out, out_type, _x == val, values[val], f"`{val}`")
        replaced[_x == val] = True

    _default = _validate_recode_default(_default, _x, out, out_type, replaced)
    out = _replace_with(
        out, out_type, ~replaced & is_not_null(_x), _default, "`_default`"
    )
    out = _replace_with(
        out,
        out_type,
        is_null(_x) | (_x == NA_character_),
        _missing,
        "`_missing`",
    )
    if out_type and not is_null(out).any():
        out = out.astype(out_type)
    return out


def _check_args(values: Mapping[Any, Any], default: Any, missing: Any) -> None:
    """Check if any replacement specified"""
    if not values and default is None and missing is None:
        raise ValueError("No replacements provided.")


@register_func(context=Context.EVAL)
def recode(
    _x: Iterable,
    *args: Any,
    _default: Any = None,
    _missing: Any = None,
    **kwargs: Any,
) -> Iterable[Any]:
    """Recode a vector, replacing elements in it

    Args:
        x: A vector to modify
        *args: and
        **kwargs: replacements
        _default: If supplied, all values not otherwise matched will be
            given this value. If not supplied and if the replacements are
            the same type as the original values in series, unmatched values
            are not changed. If not supplied and if the replacements are
            not compatible, unmatched values are replaced with NA.
        _missing: If supplied, any missing values in .x will be replaced
            by this value.

    Returns:
        The vector with values replaced
    """
    if is_scalar(_x):
        _x = [_x]

    if not isinstance(_x, numpy.ndarray):
        _x_obj = numpy.array(_x, dtype=object)  # Keep NAs
        _x = numpy.array(_x)
        if numpy.issubdtype(_x.dtype, numpy.str_):
            na_len = len(NA_character_)
            if (_x.dtype.itemsize >> 2) < na_len:  # length not enough
                _x = _x.astype(f"<U{na_len}")
            _x[is_null(_x_obj)] = NA_character_
        elif numpy.issubdtype(_x.dtype, numpy.integer):
            _x[is_null(_x_obj)] = NA_integer_

    if numpy.issubdtype(_x.dtype, numpy.number) or numpy.issubdtype(
        Array(_x[is_not_null(_x)].tolist()).dtype, numpy.number
    ):
        return _recode_numeric(
            _x, *args, _default=_default, _missing=_missing, **kwargs
        )

    return _recode_character(
        _x, *args, _default=_default, _missing=_missing, **kwargs
    )


@recode.register((Categorical, Series), context=Context.EVAL)
def _(
    _x: Iterable,
    *args: Any,
    _default: Any = None,
    _missing: Any = None,
    **kwargs: Any,
) -> Categorical:
    """Recode factors"""
    if not is_categorical(_x):  # non-categorical Series
        return recode(
            _x.values, *args, _default=_default, _missing=_missing, **kwargs
        )
    if isinstance(_x, Series):
        _x = _x.values  # get the Categorical object

    values = _args_to_recodings(*args, **kwargs)
    if not values:
        raise ValueError("No replacements provided.")

    if not all(isinstance(key, str) for key in values):
        raise ValueError(
            "Named values required for recoding factors/categoricals."
        )

    if _missing is not None:
        raise ValueError("`_missing` is not supported for factors.")

    n = len(_x)
    _check_args(values, _default, _missing)
    out = numpy.array([NA] * n, dtype=object)
    replaced = numpy.array([False] * n)
    out_type = None

    for val in values:
        if out_type is None:
            out_type = type(_get_first([values[val]]))
        out = _replace_with(
            out,
            out_type,
            # _x.categories == val,
            _x == val,
            values[val],
            f"`{val}`",
        )
        # _x may have duplicated values
        # replaced[_x.categories == val] = True
        replaced[_x == val] = True

    _default = _validate_recode_default(_default, _x, out, out_type, replaced)
    out = _replace_with(out, out_type, ~replaced, _default, "`_default`")

    if out_type is str:
        return Categorical(out)
    return out[_x.codes]


@register_func(None, context=Context.EVAL)
def recode_factor(
    _x: Iterable[Any],
    *args: Any,
    _default: Any = None,
    _missing: Any = None,
    _ordered: bool = False,
    **kwargs: Any,
) -> Iterable[Any]:
    """Recode a factor

    see recode().
    """
    values = _args_to_recodings(*args, **kwargs)
    recoded = recode(_x, values, _default=_default, _missing=_missing)

    out_type = type(_get_first(recoded))
    _default = _recode_default(_x, _default, out_type)
    all_levels = unique(
        c(
            list(values.values()),
            [] if _default is None else _default,
            [] if _missing is None else _missing,
        )
    )

    recoded_levels = (
        recoded.categories
        if isinstance(recoded, Categorical)
        else unique(recoded[pandas.notna(recoded)])
    )
    levels = intersect(
        all_levels, recoded_levels, __calling_env=CallingEnvs.REGULAR
    )

    return Categorical(recoded, categories=levels, ordered=_ordered)


recode_categorical = recode_factor



"""Provides specific contexts for datar"""
from collections import defaultdict
from enum import Enum
from typing import Any, ClassVar
from pandas.core.frame import DataFrame
from pipda.context import (
    ContextBase,
    ContextEval as ContextEvalPipda,
    ContextMixed,
    ContextPending,
    ContextSelect,
)
from .exceptions import ColumnNotExistingError


class ContextEval(ContextEvalPipda):
    """Evaluation context with used references traced"""

    name: ClassVar[str] = "eval"

    def __init__(self):
        self.used_refs = defaultdict(lambda: 0)

    def getitem(
        self, parent: Any, ref: Any, is_direct: bool = False, _attr=False
    ) -> Any:
        """Interpret f[ref]"""

        if is_direct and isinstance(ref, slice):
            # Allow f[1:3] to be interpreted as 1:3
            from .collections import Collection

            return Collection(ref)

        if is_direct:
            self.used_refs[ref] += 1
        if isinstance(parent, DataFrame):
            from .utils import df_getitem

            try:
                return df_getitem(parent, ref)
            except KeyError as keyerr:
                raise ColumnNotExistingError(str(keyerr)) from None

        return (
            super().getitem(parent, ref)
            if not _attr
            else super().getattr(parent, ref)
        )

    def getattr(self, parent: Any, ref: Any, is_direct: bool = False) -> Any:
        """Evaluate f.a"""
        return self.getitem(parent, ref, is_direct, _attr=True)

    @property
    def ref(self) -> ContextBase:
        """Defines how `item` in `f[item]` is evaluated.

        This function should return a `ContextBase` object."""
        return self


class Context(Enum):
    """Context enumerator for types of contexts"""

    UNSET = None
    PENDING = ContextPending()
    SELECT = ContextSelect()
    EVAL = ContextEval()
    MIXED = ContextMixed()
